/*
Tittle : LiFePo4_charger.c
Author : Fayaz Hassan
MCU : ATMEGA8
Date : 09-12-2020
Purpose : To charge two nos of LiFePO4 cells with automatic CC and CV modes
          and display the status on 2x16 LCD using 4bit data connection.
*/



#define F_CPU 1000000L 


#include <avr/io.h>
#include <util/delay.h>
#include <avr/eeprom.h>

#define  ON  1
#define  OFF 0

#define CC_MODE 1
#define CV_MODE 0

#define LCD_DATA_PORT	PORTD
#define LCD_DATA_REG	DDRD
#define LCD_DATA_BITS	0b11110000  


#define LCD_CTRL_PORT	PORTB
#define LCD_CTRL_REG	DDRB
#define LCD_ENABLE		0b00000001  //PB0
#define LCD_RS			0b00000010  //PB1


#define OUTPUT_PORT	PORTD
#define OUTPUT_REG  DDRD
#define OUTPUT_RELAY1  0b00000010  //PD1
#define OUTPUT_RELAY2  0b00000100  //PD2
#define OUTPUT_BLINK   0b00001000  //PB3


int MODE_BAT1 = 000;
int MODE_BAT2 = 000;
int EMF_BAT1  = 000;
int EMF_BAT2  = 000;
int EMF_MIN   = 350;
int EMF_MAX   = 360;
int Relay1status = CV_MODE; 
int Relay2status = CV_MODE; 
char TEMP_STR[10];
//==============================================================
void initAdc (  )
{

    ADCSRA |= ( (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0) ); 
	ADCSRA |= (1 << ADEN);
	ADMUX = 0b11000000; 
}
//==============================================================
int readAdc ( int channelno ) 
{
	ADMUX &= 0b11111000;
	ADMUX |= (channelno%7);  // select 0 to 7

    ADCSRA |= (1 << ADSC);

    while (ADCSRA & (1 << ADSC));

	return ADC;
    
}
//==============================================================
void initPorts ( )
{
	DDRC = 0x00; 
	DDRB = 0;
	DDRD = 0;
	LCD_DATA_REG |= ( LCD_DATA_BITS );
	LCD_CTRL_REG |= ( LCD_ENABLE | LCD_RS  );

//	STATUS_REG |= (  STATUS_HIGH |  STATUS_LOW  );

	OUTPUT_REG |= ( OUTPUT_RELAY1 | OUTPUT_RELAY2 | OUTPUT_BLINK  );
}
//=============================================================================
void enableLCD ( )
{
	

    LCD_CTRL_PORT |= LCD_ENABLE; 
    _delay_us(10);	
   LCD_CTRL_PORT &= (~LCD_ENABLE );
   _delay_us(50);		
}
//=============================================================================
void LCD_writeNibble ( unsigned char nibble , unsigned char RS)
{

	unsigned char nibl = 0; 

	if ( nibble & 0b10000000 )	nibl |= 0b00010000;
	if ( nibble & 0b01000000 )	nibl |= 0b00100000;
	if ( nibble & 0b00100000 )	nibl |= 0b01000000;
	if ( nibble & 0b00010000 )	nibl |= 0b10000000;
	if ( nibble & 0b00001000 )	nibl |= 0b00000001;
	if ( nibble & 0b00000100 )	nibl |= 0b00000010;
	if ( nibble & 0b00000010 )	nibl |= 0b00000100;
	if ( nibble & 0b00000001 )	nibl |= 0b00001000;


	LCD_DATA_PORT &= 0x0F; 

	if ( nibl <= 0b00001111)
		LCD_DATA_PORT |= ((nibl<<4)&0xF0);   
	else
		LCD_DATA_PORT |= (nibl & 0xF0);  

	if (RS)
		LCD_CTRL_PORT |= LCD_RS;
	else
		LCD_CTRL_PORT &= (~LCD_RS);

	enableLCD ( );
}
//=============================================================================
void LCD_writeCommand ( unsigned char cmd )
{
	LCD_writeNibble ( cmd & 0xF0 , 0);
	LCD_writeNibble ( cmd & 0x0F , 0);
}
//=============================================================================
void LCD_gotoRowCol ( unsigned char Row, unsigned char Col )
{
	
		
	switch (Row)
	{
	case 2:		LCD_writeCommand ( 0xC0 + Col-1 ); 	break;
	case 1:
	default:	LCD_writeCommand ( 0x80 + Col-1 );	break;
	}
	_delay_ms ( 3 );
}
//==========================================================================
void LCD_clrscr ( )
{
	LCD_writeCommand ( 0x01 );
	_delay_ms(3);
}
//=============================================================================
void LCD_init(  )
{
	unsigned char initval = 0x20; 
	
	_delay_ms ( 50 ); // > 15 ms
	LCD_writeNibble  ( 0x30 , 0);
	_delay_ms ( 20 ); // > 4.1 ms
	LCD_writeNibble ( 0x30 , 0);
	_delay_us ( 200 ); // > 37 us
	LCD_writeNibble  ( 0x30 , 0);
	_delay_ms ( 100 );
	
	LCD_writeNibble  ( 0x20 , 0 );
	_delay_ms ( 25 );

	initval |= 0b00001000; 
	initval |= 0b00000100;  
	LCD_writeCommand ( initval ); 
	_delay_ms ( 25 );

	LCD_writeCommand ( 0x0C ); 
	_delay_ms ( 25 );
	LCD_writeCommand ( 0x06 ); // increment mode
	_delay_ms ( 250 );
//	LCD_clrscr ( );
//	_delay_ms ( 10 );
}
//==================================================================
void LCD_writeChar ( unsigned char c )
{
	LCD_writeNibble ( c & 0xF0 , 1);
	LCD_writeNibble ( c & 0x0F , 1);
}
//===================================================================
void LCD_writeString ( char s[] )
{
	for (int i=0; s[i]!=0; i++)
		{
		LCD_writeChar ( s[i] );
		_delay_us ( 20 );  
		}
}
//=================================================================
char* volts2str ( int amps )
{
	int val = amps, i=0;


	val = val%1000;
	if (val>=100)
	TEMP_STR[i++] = val/100 + '0';
	TEMP_STR[i++] = '.';
	val = val%100;
	TEMP_STR[i++] = val/10 + '0';
	TEMP_STR[i++] = val%10 + '0';
	TEMP_STR[i++] = 0;
	return (TEMP_STR);

}
//==============================================================
void setRelay1 ( int status )
{

	Relay1status = status;
	if (status) //== on
		OUTPUT_PORT |= OUTPUT_RELAY1;
	else // off
		OUTPUT_PORT &= (~OUTPUT_RELAY1);

}
//===============================================================
void setRelay2 ( int status )
{
	Relay2status = status;
	if (status) //== on
		OUTPUT_PORT |= OUTPUT_RELAY2;
	else // off
		OUTPUT_PORT &= (~OUTPUT_RELAY2);

}
//===============================================================
void updateAll ( )
{

	LCD_gotoRowCol ( 1,1 );
	LCD_writeString ( "BAT1:" );

	if (Relay1status == CC_MODE && EMF_BAT1>600)
	{
		LCD_writeString ( "FULL/ABSENT" );
		setRelay1 ( CV_MODE );
	}
	else if (Relay1status == CV_MODE && EMF_BAT1>=365)
	{
		LCD_writeString ( "FULL/ABSENT" );
		setRelay1 ( CV_MODE );
	}
	else if ( EMF_BAT1<250 )
	{
		LCD_writeString ( volts2str (EMF_BAT1) );
		LCD_writeString ( "DEAD/ERR  " );
		setRelay1 ( CV_MODE );
	}
	else if ( EMF_BAT1>=345 )
	{
		LCD_writeString ( volts2str (EMF_BAT1) );
		LCD_writeString ( "V   CV " );
		setRelay1 ( CV_MODE );
	}
	else if ( EMF_BAT1<=340 )
	{
		LCD_writeString ( volts2str (EMF_BAT1) );
		LCD_writeString ( "V   CC " );
		setRelay1 ( CC_MODE );
	}
	

	LCD_gotoRowCol ( 2,1 );
	LCD_writeString ( "BAT2:" );

	if (Relay2status == CC_MODE && EMF_BAT2>600)
	{
		LCD_writeString ( "FULL/ABSENT" );
		setRelay2 ( CV_MODE );
	}
	else if (Relay2status == CV_MODE && EMF_BAT2>=365)
	{
		LCD_writeString ( "FULL/ABSENT" );
		setRelay2 ( CV_MODE );
	}
	else if ( EMF_BAT2<250 )
	{
		LCD_writeString ( volts2str (EMF_BAT2) );
		LCD_writeString ( "DEAD/ERR  " );
		setRelay2 ( CV_MODE );
	}
	else if ( EMF_BAT2>=345 )
	{
		LCD_writeString ( volts2str (EMF_BAT2) );
		LCD_writeString ( "V   CV " );
		setRelay2 ( CV_MODE );
	}
	else if ( EMF_BAT2<=340 )
	{
		LCD_writeString ( volts2str (EMF_BAT2) );
		LCD_writeString ( "V   CC " );
		setRelay2 ( CC_MODE );
	}
}
//==============================================================
void setBlinker ( int status )
{
	if (status) //== on
	{
		OUTPUT_PORT |= OUTPUT_BLINK;
	}
	else // off
	{
		OUTPUT_PORT &= (~OUTPUT_BLINK);
	}
}
//==============================================================
int main ( )
{

	_delay_ms ( 100 );
	initPorts ( );
	_delay_ms ( 100 );

	
	LCD_init(  );

	_delay_ms ( 100 );

	LCD_clrscr ( );
	LCD_writeString ( "LiFePO4 Charger" );
	LCD_gotoRowCol ( 2, 12 );
	LCD_writeString ( "-fhn" );

	initAdc (  );

	_delay_ms (100);

	while ( 1 )
	{
		setBlinker ( OFF );
		_delay_ms ( 2000 );
		setBlinker ( ON );
		_delay_ms ( 50 );
		EMF_BAT1 = readAdc ( 1 )/2;
		_delay_ms ( 50 );
		EMF_BAT2 = readAdc ( 2 )/2;
		_delay_ms ( 50);
		updateAll ( );

	}

	return 0;

} // end of main

//========================================================================================
































